
// Parte del paquete de GP
   package GP;

/**
 * Mediante la clase <B>Individuo</B>, la maquina de GP representa a
 * cada una de las instancias de arboles de funcionales de una
 * poblacion.<BR>
 *
 * Adicionalmente al arbol de composicion, el individuo encapsula las
 * calificaciones de evaluacion otorgadas al arbol en la solucion del
 * problema.<BR>
 *
 * @author Antonio Pena
 * @version jun-1997
 */
   public class Individuo implements Ordenable, Cloneable {

   /**
    * Las caracteristicas de construccion del individuo.<BR> */
      protected Especie                 especie;

   /**
    * La "rama generadora del resultado" (Result Producing
    * Branch).<BR> */
      protected Rama                    rpb;
      
   /**
    * La coleccion de estructuras de ADFs en el individuo.<BR> */
      protected ADFStru[]               adfs;

   /**
    * La calificacion de adecuacion ajustada del individuo.<BR> */
      protected float                   adjusted_fitness;

   /**
    * La calificacion de adecuacion normalizada del individuo.<BR> */
      protected float                   normalized_fitness;
      


   /**
    * Un constructor vacio.<BR> */
      public Individuo() { }

   /**
    * Construye un individuo de acuerdo a una especificacion de
    * especie.<BR>
    *
    * @param esp
    *    Es una instancia de <B>Especie</B> que contiene las
    *    caracteristicas del individuo a construir.<BR>
    * @param prof
    *    La profundidad maxima de construccion.<BR>
    * @param bal
    *    El indicador de balanceo en la rama.<BR> */
      public Individuo( Especie esp, int prof, boolean bal ) {
         this.especie = esp;

      // Crea la rama principal
         this.rpb = new Rama( esp.getTipoRPB(), prof, bal );

      // Crea los ADFs -si los hay-
         this.adfs = new ADFStru[ esp.getNADFs() ];
         for (int i=1; i<=this.adfs.length; i++)
            this.adfs[ i-1 ] =  new ADFStru( esp.getTipoADF( i ) );
      }


   /**
    * Produce una copia "profunda" del individuo.<BR> */
      public Object clone() {
         Individuo clon = new Individuo();

      // Observa que la especie no se clona. Solo se copia la referencia
         clon.especie = this.especie;

      // La rama principal
         clon.rpb = (Rama) this.rpb.clone();
         
      // Clona el arreglo de ADFs
         clon.adfs = new ADFStru[ this.adfs.length ];
         for (int i=0; i<this.adfs.length; i++)
            clon.adfs[ i ] =  (ADFStru) ((ADFStru) this.adfs[ i ]).clone();

      // Sus calificaciones
         clon.adjusted_fitness = this.adjusted_fitness;
         clon.normalized_fitness = this.normalized_fitness;

         return clon;
      }


   /**
    * Devuelve el total de nodos en la rpb y ADFs del individuo.<BR>
    * */
      public int getNNodos() {
         int           n = this.rpb.getNNodos();

         for (int i=1; i<=this.adfs.length; i++)
            n += ((ADFStru) this.adfs[i-1]).getNNodos();

         return n;
      }


   /**
    * Devuelve el indice de la Rama que contiene al nodo n_nodo.<BR>
    *
    * @return -1  si no existe tal nodo.<BR>
    *          0  si el nodo esta en la rama principal.<BR>
    *          >0 si el nodo esta en la n-esima ADF.<BR> */
      public int indiceRama( int n_nodo ) {

      // Esta en la rama principal?
         if ( (n_nodo -= this.rpb.getNNodos()) < 1 )   return 0;

      // Buscalo en alguno de las ADFs
         for (int i=0; i<this.adfs.length; i++)
            if ( (n_nodo -= this.adfs[ i ].getNNodos()) < 1 )   return i+1;

         return -1;
      }


   /**
    * Devuelve la referencia a la Rama que contiene el nodo n_nodo.<BR> */
      public Rama referenciaRama( int n_nodo ) {
         int idx_rama = this.indiceRama( n_nodo );
      
      // De acuerdo al indice de la rama que contiene al nodo
         switch ( idx_rama ) {
            case -1: return null;
            case  0: return this.rpb;
            default: return this.adfs[ idx_rama-1 ];
         }
      }


   /**
    * Devuelve una referencia al n-esimo nodo del individuo.<BR> */
      public AtomLink buscaNodo( int n_nodo ) {
         int idx_rama = this.indiceRama( n_nodo );
      
      // De acuerdo al indice de la rama que contiene al nodo
         switch ( idx_rama ) {
            case -1: return null;
            case  0: return this.rpb.buscaNodo( n_nodo );
            default: 
                  // Normaliza el numero de nodo a la ADF identificada
                     n_nodo -= this.rpb.getNNodos();
                     for (int i=1; i<idx_rama; i++)
                        n_nodo -= this.adfs[i-1].getNNodos();

                  // Ahora SI lo puedes buscar
                     return this.adfs[idx_rama-1].buscaNodo( n_nodo );
         }
      }


   /**
    * Genera una mutacion en un individuo por sustitucion de una
    * seccion de su arbol.<BR> */
      public void mutacion( Poblacion pob ) {
         Individuo mutante = (Individuo) this.clone();

      // Elige al azar el punto de mutacion
         int pm = Random.randomInt( mutante.getNNodos() );

      // Busca el punto de corte
         AtomLink corte = mutante.buscaNodo( pm );

      // Es un nodo RPB o ADF ?
         int idx_rama = mutante.indiceRama( pm );
         
      // Fabrica una nueva rama y sustituyela en su lugar de corte
         int depth;
         Funcional nuevo;
         switch ( idx_rama ) {
            case -1: return;
            case  0: TipoRama tipo_r = mutante.rpb.tipo;
                     depth = Random.randomInt(
                             1 + tipo_r.getProfMax() - corte.getProf() );
                     nuevo = new Rama( tipo_r, depth,
                                       Random.randomInt(2) == 1 ).getRaiz();
                     mutante.rpb.sustituye( corte, nuevo );
                     break;
            default: TipoADF tipo_a = mutante.adfs[ idx_rama-1 ].tipo;
                     depth = Random.randomInt(
                             1 + tipo_a.getProfMax() - corte.getProf() );
                     nuevo = new ADFStru( tipo_a, depth,
                                          Random.randomInt(2) == 1 ).getRaiz();
                     mutante.adfs[ idx_rama-1 ].sustituye( corte, nuevo );
                     break;
         }

      // Agrega al mutante a la nueva poblacion
         pob.addIndividuo( mutante );
      }


   /**
    * Obten la rama principal del individuo.<BR> */
      public Rama getRPB() { return this.rpb; }


   /**
    * Obten la calificacion de adecuacion ajustada (adjusted fitness)
    * del individuo.<BR> */
      public float getFa() { return this.adjusted_fitness; }

   /**
    * Obten la calificacion de adecuacion normalizada (normalized
    * fitness) del individuo.<BR> */
      public float getFn() { return this.normalized_fitness; }

   /**
    * Fija la calificacion de adecuacion ajustada (adjusted fitness)
    * del individuo.<BR> */
      public void setFa( float f ) { this.adjusted_fitness = f; }

   /**
    * Fija la calificacion de adecuacion normalizada (normalized
    * fitness) del individuo.<BR> */
      public void setFn( float f ) { this.normalized_fitness = f; }


   /**
    * El metodo <B>cruzamiento</B> permite
    * producir la cruza de dos individuos.<BR>
    *
    * @param ind
    *    Es el segundo padre en el cruzamiento.<BR> */
      public void cruzamiento( Individuo ind, Poblacion pob ) {
         Rama      rama_a,  rama_b;
         Individuo copia_a, copia_b;


      // Copia a los dos individuos
         copia_a = (Individuo) this.clone();
         copia_b = (Individuo)  ind.clone();

      // La informacion sobre los cortes
         AtomLink liga_a, liga_b;
         int corte_a, corte_b, above_a, above_b, below_a, below_b;
         corte_a = corte_b = above_a = above_b = below_a = below_b = 0;

      // Busca un par de cortes validos
         do {

         // El sitio de corte en 'a'
            corte_a = Random.randomInt( copia_a.getNNodos() );
            int nrama_a = copia_a.indiceRama( corte_a );
            rama_a = (Rama) copia_a.referenciaRama( corte_a );
            liga_a = copia_a.buscaNodo( corte_a );
            above_a = liga_a.getProf() - 1;
            if ( liga_a.getIHijo() == 0 )
               below_a = liga_a.getPadre().getProf();
            else
               below_a = liga_a.getHijo().getProf();
         

         // Obten un punto de corte en 'b'
            switch ( nrama_a ) {
               case -1: return;
               case  0: corte_b = Random.randomInt( copia_b.rpb.getNNodos() );
                        rama_b =  copia_b.rpb;
                        break;
               default: corte_b = Random.randomInt(
                           copia_b.adfs[ nrama_a-1 ].getNNodos() );
                        rama_b =  copia_b.adfs[ nrama_a-1 ];
                        break;
            }

         // La liga de corte en 'b'
            liga_b = rama_b.buscaNodo( corte_b );
            above_b = liga_b.getProf() - 1;
            if ( liga_b.getIHijo() == 0 )
               below_b = liga_b.getPadre().getProf();
            else
               below_b = liga_b.getHijo().getProf();

      // Valida la profundidad de los resultados
         } while ( above_a + below_b > rama_a.tipo.getProfMax()
                || above_b + below_a > rama_b.tipo.getProfMax() );


      // Los nuevos subarboles
         Funcional nuevo_a = liga_a.getIHijo() > 0
                                ? liga_a.getHijo()
                                : liga_a.getPadre();

      // Sustituye en A y B
         Funcional nuevo_b = rama_b.sustituye( liga_b, nuevo_a );
         rama_a.sustituye( liga_a, nuevo_b );


      // Agrega los nuevos individuos a la poblacion
         pob.addIndividuo( copia_a );
         pob.addIndividuo( copia_b );
      }


   // Devuelve la referencia al ADFStru de nombre 'name'
      public ADFStru getADFStru( String name ) {
         ADFStru ref = null;
         
      // Busca en todo el vector de ADF's
         for (int i=0; i<this.adfs.length; i++) {
         
         // Devuelve el ADF si el nombre es correcto
            if ( name.equals(((ADFStru) adfs[i]).getNombre()) )
               ref = (ADFStru) adfs[ i ];
         }

         return ref;
      }


   /**
    * Implementacion del metodo <B>orden</B> de la clase
    * <B>Ordenable</B>.<BR>
    *
    * @param cmp
    *    Un segundo <B>Individuo</B> con el cual comparar
    *    calificaciones de adecuacion normalizadas (normalized
    *    fitness).<BR>
    *
    * @return <B>x</B> < 0 si <B>this</B>  < <B>cmp</B>.<BR>
    *         <B>x</B> = 0 si <B>this</B> == <B>cmp</B>.<BR>
    *         <B>x</B> > 0 si <B>this</B>  > <B>cmp</B>.<BR> */
      public int orden( Ordenable cmp ) {

         if ( this.normalized_fitness
           == ((Individuo) cmp).normalized_fitness ) {
            return this.getNNodos() - ((Individuo) cmp).getNNodos();
         } else
         return
            this.normalized_fitness < ((Individuo) cmp).normalized_fitness
               ? -1 : +1;
      }


   /**
    * Devuelve una representacion en <B>String</B> del individuo.<BR>
    * */
      public String toString() {
         String str = new String( "Individuo :" );
         
         str = str.concat("\nrpb -> ");
         str = str.concat(this.rpb.getRaiz().toString());

         for (int i=0; i<this.adfs.length; i++) {
            str = str.concat("\nadf "+i+" -> ");
            str = str.concat(
                      ((ADFStru) this.adfs[ i ]).getRaiz().toString());
         }
         
         return str.toString();
      }
   }

